Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq.
Require Import pred prelude idynamic pcm unionmap heap coding feap.
Require Import state zigzag world inject rely.
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

(* Refinement structure. Finee world F refines coarser world C if there *)
(* is a partial bijection between the coherent worlds, with a number of *)
(* simulation properties. The refinement structure captures the properties *)

Module Refinement.
Section Refinement.
Variables (seedT : Type) (C F : mod).
(* seedT is the type of extra seed information used for refining. The *)
(* seed information is a ghost that uniquely determines the refined state *)
(* from the coarse one. Thus, the ghost can be used as a breadcrumb *)
(* during operational semantics to record the state of evaluation wrapped *)
(* by a Hide constructor. Whenever the program makes one step, the state *)
(* is changed and recorded under Hide, so that the continuation can *)
(* proceed to step from the new state. In principle, we could have just *)
(* recorded the refined state under Hide, rather than introduce the seed *)
(* ghosts. However, that makes the program specific to one particular *)
(* state (the one recorded under Hide), whereas we want program to be *)
(* able to start in many different states, depending on the history of *)
(* execution. Thus, the seeds a programming crutch that identify a class *)
(* of different refined states, depending on the supplied coarse state. *)

(* 
Since seeds appear under Hide, the initial one always has to be
explicitly supplied by the programmer. This may sound annoying at
first, though I hope it won't be too demanding in practice. The
initial seed should usually be some constant (e.g., a unit of a
concurroid), rather than a parameter that dependents on the state
(thought there certainly are such examples).

Be it as it may, I have considered a version of a refinement structure
which doesn't require seeds. In that setting, one has to require
additional properties of the refinement STSs and their programs. In
particular, since Hide doesn't record the refined state in which the
hidden program executes, and this refined state is not uniquely
determined by the coarse state, we have to demand that the
nondeterministic choice of possible refinements doesn't matter. That
is, denoting by s1' ~ s2' the property that s1' and s2' are both
refinement a common s, we need an *erasure* property that
  
s1' ~ s2' -> tstep s1' p t1' q1 -> 
             tstep s2' p t2' q2 -> t1' ~ t2' 
(and also some erasure between q1 and q2 which lifts ~)

Now, this is not a property of STS's, but of trees p, q1 and q2 which
live in the fine-grained STS. Thus, it's not something encodable in
the refinement structure, but one has to bake into trees, starting
from atomic actions and onward to the compound trees. Thus, we need
not only a structure of actions living at an STS F, but also "erasing"
to some other STS C. Moreover, one has to build a whole tower of
erasures, in case C is further coarsened into C', etc. Such a setup is
obviously abstraction-breaking; if we want to coarsen a program, we
need to know how it has been built intensionally. We need to know
which atomic actions it uses in order to prove that they can be
coarsened to C'.

To remedy the issue, one needs to associate STSs with the actions that
its programs are allowed to use. Then refinement structure can at
least check if each of the actions in F erase to some action in C in
the above sense (or maybe map to idle in C). But this is a bit
different from how we envisioned things. What if we don't want C to
have any atomic actions associated with it, because we only want to
use it for refinement, not for programming? E.g., if F is an
implementation of a fine-grained stack and C is a coarse-grained one,
then C need not have any atomic actions of adding/removing to the
stack. These operations are programs in C, obtained by refinement from
programs in F, but they are not actions in the sense that their
granularity is clearly not a single memory operation. 

This seems to imply that we don't really want to keep track of atomic
actions in an STS, but of all the components that make up the internal
transition. In other words, we want a number of internal transitions
instead of just one -- as is now the case -- which is a big
disjunction of all the actual moves the STS allows. In this setting,
it seems we can define refinement as a property of STSs, rather than
trees, but this requires a much more serious change to the code base,
so let's do it only if we have to, i.e., only if writing the initial
seeds in programs becomes unbearable.

For this to work the Priv module has to be modified, so that it's
transitions all correspond to something deterministic, e.g.  a change
to a single pointer, etc. Otherwise, we won't be able to prove
erasure.

Also, one may worry that transitions are not quite like actions, as
they don't return a result value. Thus, why is it OK to replace
actions by transitions in this case? In particular, what can we say
about q1 and q2 in the above erasure equation, and how may we go about
proving it?

I think we should demant that q1 = q2. To see why, consider the base
case of atomic actions, which is the only interesting one, I think.
In that case, we're running the same action in two different states,
albeit which both erase to the same state. We know that the value
returned will be the same, simply because the two initial states
heap-erase to the same heap, and so by erasure property, the values
must be equal.

*)

Structure mixin_of (phi : seedT -> reltp) := Mixin {
  (* phi is a partial function; refining states are uniquely determined *)
  (* but some states in F need not refine into anything in C. *)
  (* E.g. we can't create a lock out of a pointer x, if x doesn't exist. *)
  (* This is an adequacy property for the seeds, needed so that we can use *)
  (* the seed in the operational semantics as a shorthand for the refined *)
  (* state *)
  _ : forall g s s1' s2', phi g s s1' -> phi g s s2' -> s1' = s2';
  (* refine is injective, i.e., a refinement has a unique erasure *)
  (* in the coarse world and a unique seed *)
  _ : forall g1 s1 g2 s2 s',   
        phi g1 s1 s' -> phi g2 s2 s' -> (g1, s1) = (g2, s2);
  (* refine works only on coherent states *)
  _ : forall g s s', phi g s s' -> 
        s \In Mod.coh C /\ s' \In Mod.coh F;
  (* The two requirements below are commutative diagrams expressing simulation *) 
  _ : forall g s1 s1' s2, 
        phi g s1 s1' -> mrely C s1 s2 -> 
        exists s2', phi g s2 s2' /\ menv_steps F s1' s2';
  (* abstracting back preserves guars; like a simulation property *)
  (* the coarse module W (which has more states) simulates the finer W' *)
  _ : forall g1 s1 s1' s2', 
        phi g1 s1 s1' -> Mod.internal F s1' s2' -> 
        exists g2 s2, phi g2 s2 s2' /\ Mod.internal C s1 s2;
  (* refinement realigns loc and shr, but doesn't move things into/outof env *)
  (* this seems not provable from zagzig, which merely states *)
  (* that refine doesn't move things out of env, but not that they are unchanged *)
  _ : forall g1 s1 s1' g2 s2 s2', 
      phi g1 s1 s1' -> phi g2 s2 s2' -> 
      (env s1 = env s2 <-> env s1' = env s2');
  (* zagzig *)
  (* every p extension to env, is (uniquely?) refined into q *)
  (* moreover, refining direction takes stuff from loc and moves to shr *)
  (* but doesn't remove stuff from env *)
  (* the oposite direction takes from shr and moves to loc *) 
  (* that q is unique can probably be infered from functionality *)
  (* the statement of the lemma is kinda of interesting; in the second part *)
  (* we say we only care to invert q's which have already been obtained from p *)
  (* in practice, obtaining q can only be done by refining with a Unit feap *)
  (* so, the statement says that we can invert back only those q which are obtained *)
  (* by refining with unit. *)
  (* if we wanted to be more general, and allow arbitrary q's *)
  (* we will have to also existentially produce new g (in addition to s1' *)
  (* since we by zigzagging q in the opposite way, we will be zigzaging *)
  (* the l ghosts in the oposite way too *)
  (* be it as it may; let's go with this statement for now *)
  _ : forall p g s s', 
        phi g (s \> p) s' -> 
        exists q s1', 
          [/\ s' = s1' \> q, phi g (s \< p) (s1' \< q) & 
              forall g s s', phi g s (s' \< q) ->
              exists s1, s = s1 \< p /\ phi g (s1 \> p) (s' \> q)];
  (* refinement only deals with ghosts *)
  _ : forall g s s', phi g s s' -> flats s = flats s'}.

Section ClassDef.

Notation class_of := mixin_of (only parsing). 
Notation base := (seedT -> reltp).

Structure refinement : Type := Pack {sort : base; _ : class_of sort; _ : base}.
Local Coercion sort : refinement >-> Funclass.

Variables (T : base) (cT : refinement).
Definition class := let: Pack _ c _ as cT' := cT return class_of cT' in c.

Definition pack c := @Pack T c T.
Definition clone := fun c & T = cT & phant_id (pack c) cT => pack c.

End ClassDef.
End Refinement.

Module Exports.
Coercion sort : refinement >-> Funclass.
Notation refinement := refinement.
Notation RefinementMixin := Mixin.
Notation Refinement T m := (@pack _ _ _ T m).

(*
Notation "[ 'refinementMixin' 'of' T ]" := (class _ : mixin_of T)
  (at level 0, format "[ 'refinementMixin'  'of'  T ]") : form_scope.
Notation "[ 'refinement' 'of' T 'for' C ]" := (@clone T C _ (erefl _) id)
  (at level 0, format "[ 'refinement'  'of'  T  'for'  C ]") : form_scope.
Notation "[ 'refinement' 'of' T ]" := (@clone T _ _ (erefl _) id)
  (at level 0, format "[ 'refinement'  'of'  T ]") : form_scope.
*)

Section Repack.
Variables (T : Type) (C F : mod) (phi : refinement T C F).

Lemma rf_func g s s1' s2' : phi g s s1' -> phi g s s2' -> s1' = s2'.
Proof. by case: phi=>q [rf inj coh rp gp ep zg fl xx]; apply: rf. Qed.

Lemma rf_inj g1 s1 g2 s2 s' : phi g1 s1 s' -> phi g2 s2 s' -> (g1, s1) = (g2, s2).
Proof. by case: phi=>q [rf inj coh rp gp ep zg fl xx]; apply: inj. Qed.

Lemma rf_coh g s s' : phi g s s' -> s \In Mod.coh C /\ s' \In Mod.coh F.
Proof. by case: phi=>q [rf inj coh rp gp ep zg fl xx]; apply: coh. Qed.

Lemma rf_mrely_sim g s1 s1' s2 : 
        phi g s1 s1' -> mrely C s1 s2 -> 
        exists s2', phi g s2 s2' /\ menv_steps F s1' s2'.
Proof. by case: phi=>q [rf inj coh rp gp ep zg fl xx]; apply: rp. Qed.

Lemma rf_guar_sim g1 s1 s1' s2' :
        phi g1 s1 s1' -> Mod.internal F s1' s2' -> 
        exists g2 s2, phi g2 s2 s2' /\ Mod.internal C s1 s2.
Proof. by case: phi=>q [rf inj coh rp gp ep zg fl xx]; apply: gp. Qed.

Lemma rf_env g1 s1 s1' g2 s2 s2' :
        phi g1 s1 s1' -> phi g2 s2 s2' -> 
        (env s1 = env s2 <-> env s1' = env s2').
Proof. by case: phi=>q [rf inj coh rp gp ep zg fl xx]; apply: ep. Qed.

Lemma rf_zagzig p g s s' :
        phi g (s \> p) s' -> 
        exists q s1', 
          [/\ s' = s1' \> q, phi g (s \< p) (s1' \< q) & 
              forall g s s', phi g s (s' \< q) ->
              exists s1, s = s1 \< p /\ phi g (s1 \> p) (s' \> q)].
Proof. by case: phi=>q [rf inj coh rp gp ep zg fl xx]; apply: zg. Qed.

Lemma rf_flats g s s' : phi g s s' -> flats s = flats s'. 
Proof. by case: phi=>q [rf inj coh rp gp ep zg fl xx]; apply: fl. Qed.

End Repack.

Prenex Implicits rf_func rf_inj rf_coh rf_mrely_sim rf_guar_sim rf_env rf_zagzig rf_flats. 

End Exports. 

End Refinement.

Export Refinement.Exports.

(* Steps by environment in the coarse world, don't change the refining
ghost gE. This is so because gE is a part of the refined world, so
it's out of reach to the stepping in the coarse world. *)

Lemma rf_menvs T C F (phi : refinement T C F) i m (gL : T) i' : 
        phi gL i i' -> menv_steps C i m -> 
        exists m', phi gL m m' /\ menv_steps F i' m'.
Proof.
move=>Rfi [n E].
elim: n i m gL i' E Rfi=>[|n IH] i m gL i' /= E Rfi.
- case: E=><-{m} C'; exists i'; split=>//.
  by apply: menvs_refl; case/rf_coh: Rfi. 
case: E=>h [Rih] E.
case: (rf_mrely_sim Rfi Rih)=>h' [Rfh] Rih'. 
case: (IH _ _ _ _ E Rfh)=>m' [Rfm] E'. 
by exists m'; split=>//; apply: menvs_trans E'.
Qed.

(* two derived legacy properties; they used to be part of refinement *)
(* but then i saw they can be proved out of the rest *)

(* I used to have that refinement function f is surjective *)
(* I.e., on all coherent states s2' in V, we can find g2,s2 in U such that f g2 s2 s2' *)
(* But that doesn't work in practice on the instances. In practice f is surjective *)
(* only if s2'.other is unit. A way to say that without directly invoking *)
(* the unit (which seems impossible in the whole setting) is that s2' *)
(* is obtained by stepping from some other s1' which is in the range of f *)
(* Then the statement simply becomes part of the bisimulation property *)

Lemma rf_guarsur T C F (phi : refinement T C F) g1 s1 s1' s2' : 
        phi g1 s1 s1' -> Mod.internal F s1' s2' ->
        exists g2 s2, phi g2 s2 s2'.
Proof. by move=>H /(rf_guar_sim H) [g2][s2][H']; exists g2, s2. Qed.

(* and then the simulation proper *)
Lemma rf_guar T C F (phi : refinement T C F) g1 s1 s1' g2 s2 s2' :
        phi g1 s1 s1' -> phi g2 s2 s2' -> 
        Mod.internal F s1' s2' -> Mod.internal C s1 s2. 
Proof. by move=>H1 H2 /(rf_guar_sim H1) [g3][s3][] /(rf_inj H2) [_ ->]. Qed.


