Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq path.
Require Import pred prelude idynamic domain pcm ordtype unionmap heap coding feap.
Require Import state zigzag world rely inject atomic schedule refine hide. 
Require Import privstate process always model lemmas stsep.
Set Implicit Arguments. Unset Strict Implicit. Import Prenex Implicits.

(******************************************************************* 
*********  Verification of a concurrent spanning-tree algorithm *****
*******************************************************************)

Section Definitions.

(* Graph definition is parametrized by the type of nodes, which need
   to ordered to simplify the comparison. *)

Variable node : ordType. 


(* Sets are represented as partial maps from set elements to the
   elements of the unit type. *)

Notation node_set := (union_map node unitSet).

(* Edges are functions, mapping nodes to sets (predicates) of
   nodes. *)

Variable edge : node -> pred node.

(* Front of a set of nodes t with respect to the edge-induced topology
   is *at least* a set of nodes, which are immediately reachable from
   t via edge. *)

Definition front (t t' : node_set) :=
  {subset dom t <= dom t'} /\ forall x y, x \in dom t -> edge x y -> y \in dom t'.

(* Set of nodes t is maximal in the (implicit) topology edge if it can
   be considered as its own front. *)

Definition maximal t := front t t.

(* Tree rooted at x is a subgraph that has unique paths from x. *)
Definition tree (x : node) (t : node_set) := 
  [/\ x \in dom t & forall y, y \in dom t -> 
                    exists !p, [/\ y = last x p, path edge x p & {subset p <= dom t}]].

(**********************************************)
(* Some uselful lemmas about trees and graphs *)
(**********************************************)

(* A tree has no cycles *)
Lemma tree_cycle x y t p : 
 tree x t -> {subset p <= dom t} -> y \in dom t -> ~~ cycle edge (y :: p).
Proof.
case=>_ T Sp Dy; apply/negP => /= Py.
case: (T y Dy) => p' [[Exy Px Sp'] Uq].
suff : p' = p' ++ rcons p y.
- by elim: {Exy Px Sp' Uq Sp Py} p'=>[|x' p' IH]; [case: p | case].
apply: Uq; rewrite last_cat last_rcons cat_path Px -Exy; split=>//.
move=>z; rewrite mem_cat mem_rcons inE.
by case/or3P; [apply: Sp' | move/eqP=>-> | apply: Sp].
Qed.

(* We can never "go outside" of a maximal t. *)
Lemma max_path t x p : 
        maximal t -> path edge x p -> x \in dom t -> {subset x :: p <= dom t}.
Proof.
case=>_ M Px Dx; elim: p x Dx Px=>[|y p IH] x Dx //=.
- by move=>_ z; rewrite inE; move/eqP => ->.
case/andP => /(M _ _ Dx) Dy Py z; rewrite inE.
by case/orP; [move/eqP=>-> | apply: IH Dy Py z].
Qed.

(* If there is an edge from x to y, and y is a root of a tree t, then
   x is not in t. *)
Lemma tree_edge x y t : edge x y -> tree y t -> x \notin dom t. 
Proof.
move=>H T; case: (T)=>D Tx; apply/negP => /Tx [p][Ex _]; case: Ex=>Ex P S.
by move: (tree_cycle T S D); rewrite /= rcons_path P -Ex H.
Qed.

(* Three basic front lemmas. *)
Lemma frontPt x t : {subset (edge x) <= dom t} -> x \in dom t -> front (#x) t.
Proof. by move=>E Dx; split=>[u|u v]; rewrite um_domPt inE=>/eqP <- // /E. Qed.

Lemma frontUn t1 t2 t : front t1 t -> front t2 t -> front (t1 \+ t2) t.
Proof.
case=>S1 F1 [S2 F2]; split=>[z|x y]; rewrite domUn; move/andP/proj2.
- by case/orP; [apply: S1 | apply: S2].
by case/orP; [apply: F1 | apply: F2].
Qed.

Lemma front_leq t t1 t2 : {subset dom t1 <= dom t2} -> front t t1 -> front t t2.
Proof.
move=>S [S1 F1]; split; first by move=>z /S1 /S.
by move=>x y Dx Ex; apply: S; apply: F1 Dx Ex.
Qed.


(* Maximality and treeness lemmas depend on the out-degree of node x
   with respect to the topology induced by edge. *)

(* - when x has no children *)

Lemma max0 x : edge x =1 pred0 -> maximal (#x).
Proof.
move=>B; split=>// u v; rewrite um_domPt inE. by move/eqP => <-{u}; rewrite B. 
Qed.

Lemma tree0 x : edge x =1 pred0 -> tree x (#x). 
Proof.
move=>B; split=>[|y]; rewrite um_domPt inE //. 
move/eqP => <-{y}; exists [::]; split=>//. by case=>//= u p []; rewrite B. 
Qed.

(* - when x has one child *)

Lemma max1 x y ty : 
edge x =1 pred1 y -> y \in dom ty -> maximal ty -> maximal (#x \+ ty). 
Proof.
move=>B Dy [_ My]; split=>// u v.  
rewrite !um_domPtUn !inE; case/andP=>->. 
case/orP; last by move=>Du /(My _ _ Du) ->; rewrite orbT.
by move/eqP=><-{u}; rewrite B; move/eqP=>->; rewrite Dy orbT.
Qed. 

Lemma tree1 x y ty : 
 edge x =1 pred1 y -> tree y ty -> maximal ty -> tree x (#x \+ ty). 
Proof.
move=>B [Dy Ty] M. 
have Dx : x \notin dom ty by rewrite (@tree_edge x y ty) // B inE.
split; first by rewrite um_domPtUnE um_validPtUn (dom_valid Dy). 
move=>u; rewrite um_domPtUn inE; case/andP=>Vx.
case/orP.
- move/eqP=><-{u}; exists [::]; split=>//.
  case=>//= u p [Exu /andP [Hu Pu] _].   
  rewrite Exu (max_path _ _ _ (mem_last _ p)) // in Dx.  
  by rewrite B inE in Hu; rewrite (eqP Hu).
move=>Du; case: (Ty _ Du)=>p [][Eyu Py Sp] Uq.
exists (y::p); split=>/=. 
- rewrite B inE eq_refl Py; split=>// z; rewrite inE => X.
  have Y : z \in dom ty by case: eqP X=>[->//|_]; apply: Sp. 
  by rewrite um_domPtUn inE Vx Y orbT.
case=>[|xx p'][/= Exu]; first by rewrite -Exu Du in Dx.
rewrite B inE; case/andP=>H Px _; rewrite {xx H}(eqP H) in Exu Px *.
rewrite (Uq p') //; split=>// z X.
by apply: (max_path M Px Dy); rewrite inE X orbT.
Qed.

(* - when x has two children *)

Lemma max2 x y1 y2 ty1 ty2 : 
  edge x =1 pred2 y1 y2 ->
  y1 \in dom ty1 -> maximal ty1 -> y2 \in dom ty2 -> maximal ty2 ->
  maximal (#x \+ (ty1 \+ ty2)). 
Proof.
move=>B Dy1 [_ My1] Dy2 [_ My2]; split=>// u v; rewrite um_domPtUn inE.
case/andP=>Vx; move: (validR Vx)=>V.
have [Dx1 Dx2] : x \notin dom ty1 /\ x \notin dom ty2.
- by rewrite um_validPtUn domUn inE V /= negb_or in Vx; case/andP: Vx.
rewrite domUn inE V /=; case/orP; last first.
- case/orP=>Dy; [move/(My1 u v Dy) | move/(My2 u v Dy)] => X;
  by rewrite um_domPtUn inE Vx domUn inE V X !orbT.
move/eqP=><-{u}; rewrite B !inE; case/orP=>/eqP ->{v}; 
by rewrite um_domPtUn inE Vx domUn inE V ?Dy1 ?Dy2 !orbT.
Qed.

Lemma tree2 x y1 y2 ty1 ty2 : 
  edge x =1 pred2 y1 y2 -> tree y1 ty1 -> maximal ty1 ->
  tree y2 ty2 -> maximal ty2 -> valid (ty1 \+ ty2) -> tree x (#x \+ (ty1 \+ ty2)).
Proof.
move=>B [Dy1 Ty1] My1 [Dy2 Ty2] My2 V.
have [Dx1 Dx2] : x \notin dom ty1 /\ x \notin dom ty2.
- rewrite (@tree_edge x y1) // ?B ?inE ?eq_refl //. 
  by rewrite (@tree_edge x y2) // ?B ?inE ?eq_refl // orbT. 
split; first by rewrite um_domPtUnE um_validPtUn domUn inE V negb_or Dx1 Dx2.
move=>u; rewrite um_domPtUn inE; case/andP=>Vx. 
case/orP.
- move/eqP=><-{u}; exists [::]; split=>//.
  case=>//= y p [Exy /andP [H Px] _].
  rewrite B !inE in H; case/orP: H Exy Px=>/eqP ->{y} Exy Px.
  - by rewrite Exy (max_path _ _ _ (mem_last _ p)) in Dx1.
  by rewrite Exy (max_path _ _ _ (mem_last _ p)) in Dx2.
rewrite domUn inE V /=.
wlog : y1 y2 ty1 ty2 B Dy1 Dy2 My1 My2 Ty1 Ty2 Dx1 Dx2 V Vx / u \in dom ty1.
- move=>G; case/orP=>H; [|rewrite (joinC ty1) in Vx V *].
  - by apply: (G y1 y2)=>//; rewrite H. 
  by apply: (G y2 y1)=>//; [move=>z; rewrite B !inE orbC | rewrite H].
move=>Du1 _; case: (Ty1 u Du1)=>p [][Ey1 Py1 Sy1] Uq.
have Du2 : u \notin dom ty2 by case: validUn V Du1=>// _ _ L _ /L /negbTE => ->.
exists (y1 :: p); split=>/=. 
- rewrite B inE eq_refl; split=>// z; rewrite inE => X.
  have Y : z \in dom ty1 by case: eqP X=>[->//|_]; apply: Sy1.
  by rewrite um_domPtUn inE Vx domUn inE V /= Y orbT.
case=>[|xx p'][/= Exy]; first by rewrite -Exy Du1 in Dx1.
rewrite B !inE; case/andP=>H Px _. 
case/orP: H Exy Px=>/eqP ->{xx} Exy Px; last first.
- by rewrite Exy (max_path My2 Px Dy2 (mem_last _ p')) in Du2.
rewrite (Uq p') //; split=>// z X.
by apply: (max_path My1 Px Dy1); rewrite inE X orbT.
Qed.

End Definitions.

(***********************************)
(* Now we connect heaps and graphs *)
(***********************************)

(* The graphs we care about have a boolean marker b at each node and
   two edges xl, xr. a heap that represents a graph stores nodes as
   pointers. *)

Definition graph (h : heap) := 
  valid h /\ 
  forall x, x \in dom h -> 
    exists (b : bool) (xl xr : ptr),
      h = x :-> (b, xl, xr) \+ free x h /\
      {subset [:: xl; xr] <= [predU pred1 null & dom h]}. 

Local Notation ptr_set := (@union_map [ordType of ptr] unitSet).

Section GraphsHeaps.
Variables (h : heap) (g : graph h).

Lemma contents_pfx (x : ptr) v : find x h = Some v -> idyn_tp v = prod (prod bool ptr) ptr. 
Proof. 
move=>F; case/find_some/(proj2 g): F (F)=>b [xl][xr][E] _.
by move: (proj1 g); rewrite E => V; rewrite hfindPtUn //; case=><-.
Qed.

(* the contents of node x; that is, the mark bit and edges of x *)

Definition contents (x : ptr) : bool * ptr * ptr := 
  match find x h as f return _ = f -> _ with
    Some v => fun epf => icoerce id (idyn_val v) (contents_pfx epf)
  | None => fun epf => (false, null, null)
  end (erefl _).

CoInductive contents_spec (x : ptr) : bool -> bool * ptr * ptr -> Type := 
| has_some (b : bool) (xl xr : ptr) of 
    h = x :-> (b, xl, xr) \+ free x h & valid h & 
    x \in dom h & {subset [:: xl; xr] <= [predU pred1 null & dom h]} :
    contents_spec x true (b, xl, xr)
| has_none of x \notin dom h : contents_spec x false (false, null, null). 

Lemma edgeP x : contents_spec x (x \in dom h) (contents x). 
Proof.
case: (g) => V G; case Dx : (x \in dom h); last first.
- case: {G} dom_find (Dx)=>// N _.
  rewrite /contents; move: (@contents_pfx x); rewrite N /= => _.
  by apply: has_none; rewrite Dx. 
case E: (contents x)=>[[b xl] xr].
suff [Ex S] : h = x :-> (b, xl, xr) \+ free x h /\
              {subset [:: xl; xr] <= [predU pred1 null & dom h]}. 
- by apply: has_some=>//; rewrite -Ex.
case/G: {G} Dx=>b' [xl'][xr'][E'] S.
move: E; rewrite /contents; move: (@contents_pfx x).
rewrite E' hfindPtUn -E' // => Ex; rewrite !ieqc /= {Ex}. 
by case=><-<-<-.
Qed.

(* the edge relation of a graph *)

Definition edge x := 
  [pred y | [&& x \in dom h, y != null & 
    let: (_, xl, xr) := contents x in y \in pred2 xl xr]].

(* graph is connected from x if it has a path from x to every other node *)
Definition connected x :=
  forall y, y \in dom h -> exists p, path edge x p /\ last x p = y.

End GraphsHeaps.

Notation mark g x := ((@contents _ g x).1.1).
Notation edgl g x := ((@contents _ g x).1.2).
Notation edgr g x := ((@contents _ g x).2).

(* restating the graph properties in terms of booleans *)

Lemma validG h (g : graph h) : valid h.
Proof. by case: g. Qed.

Lemma edgeE h (g : graph h) h' x c : h = x :-> c \+ h' -> contents g x = c.
Proof.
case: edgeP; first by move=>b xl xr -> V Dx _ /(hcancelV V) []. 
by move=>Dx E; rewrite E hdomPtUn inE eq_refl /= andbT -E (validG g) in Dx.
Qed.

Implicit Arguments edgeE [h g h' x c].

Lemma edgeG h (g : graph h) x : 
        {subset [:: edgl g x; edgr g x] <= [predU pred1 null & dom h]}.
Proof. by case: edgeP=>//= _ z; rewrite !inE orbb=>->. Qed.

Lemma mark_dom h (g : graph h) x : mark g x -> x \in dom h. 
Proof. by case: edgeP. Qed.

Lemma eqG h1 (g1 : graph h1) h2 (g2 : graph h2) x : 
        h1 = h2 -> contents g1 x = contents g2 x.
Proof. by move=>E; rewrite -E in g2 *; rewrite (proof_irrelevance g1 g2). Qed.


(* Functions for computing contents of updated graphs. *)

Definition mark_node h (g : graph h) x := 
  if x \in dom h then x :-> (true, edgl g x, edgr g x) \+ free x h else h.

(* Graph with a mrked node is still a graph. *)
Lemma mark_nodeG h (g : graph h) x : graph (mark_node g x).
Proof.
rewrite /mark_node; case: edgeP=>//= b xl xr E V Dx S.
split=>[|y]; first by rewrite {1}E !hvalidPtUn in V *.
rewrite hdomPtUn inE; case/andP=>Vt.
case N : (x == y)=>/=.
- rewrite -(eqP N) hfreePtUn //; exists true, xl, xr; split=>//.
  by move=>z /S; rewrite {1}E !inE /= !inE /= !hdomPtUn -E V Vt.
move=>Dy; have {Dy} : y \in dom h by rewrite E domUn inE -E V Dy orbT.
case: edgeP=>// b' xl' xr' E' _ Dy S' _; rewrite {1 2}E in V E'.
move/(hcancel2V V): E'; rewrite N; case=>Exy Ex _.
exists b', xl', xr'. 
rewrite {1}Ex joinCA freeUnL; last by rewrite hdomPt inE N andbF.
rewrite Exy; split=>// z /S'. 
by rewrite {1}E !inE /= !inE /= !hdomPtUn !inE !hvalidPtUn.
Qed.

Lemma mark_domE h (g : graph h) x : dom (mark_node g x) =i dom h.
Proof.
rewrite /mark_node => y; case: edgeP=>[b xl xr E _ _ _|//]. 
by rewrite {2}E !hdomPtUn !inE !hvalidPtUn.
Qed.

Lemma mark_nodeE h (g : graph h) x (g' : graph (mark_node g x)) y : 
        contents g' y = 
        if (x \in dom h) && (x == y) then (true, edgl g x, edgr g x)
        else contents g y.
Proof.
rewrite /mark_node; case: edgeP; last first.
- case: edgeP=>/=; last by case: edgeP.
  move=>b xl xr E V Dx _; rewrite E in V Dx; case: ifP=>N. 
  - by rewrite -(eqP N) !hdomPtUn !inE !hvalidPtUn in Dx *; rewrite Dx.
  move=>M; suff: y \notin dom h by case: edgeP.
  by rewrite E; rewrite !hdomPtUn !inE !hvalidPtUn in M *. 
move=>b xl xr; case: edgeP=>/=; last by move=>_ /edgeE ->.
move=>b1 xl1 xr1 Ex Vx Dx _ E Vy Dy _.
move/(hcancel2V Vy): E; case: ifP=>N [// _ E _].
by rewrite E joinCA in Ex; rewrite (edgeE Ex).
Qed.

Inductive side := Left | Right. 

Definition null_edge h (g : graph h) (c : side) x := 
  if mark g x then 
    x :-> (if c is Left then (true, null, edgr g x) 
           else (true, edgl g x, null)) \+ free x h
  else h.

(* Graph with a removed edge is still agraph. *)
Lemma null_edgeG h (g : graph h) c x : graph (null_edge g c x).
Proof.
rewrite /null_edge; case: edgeP=>//=; case=>// xl xr E V Dx S.
split=>[|y]; first by rewrite {1}E !hvalidPtUn in V *.
rewrite hdomPtUn inE; case/andP=>Vt.
case N : (x == y)=>/=.
- rewrite -(eqP N) hfreePtUn //. 
  exists true, (if c is Left then null else xl), 
               (if c is Left then xr else null).
  split=>[|z]; first by case: c Vt.
  move: {S} (S z); rewrite {1}E !inE /= !inE !hdomPtUn !inE -E V Vt /=.
  by case: c Vt=>Vt S; case/orP=>X; rewrite ?X //; apply: S; rewrite X // orbT.
move=>Dy; have {Dy} : y \in dom h by rewrite E domUn inE -E V Dy orbT.
case: edgeP=>// b' xl' xr' E' _ Dy S' _; rewrite {1 2}E in V E'.
move/(hcancel2V V): E'; rewrite N; case=>Exy Ex _.
exists b', xl', xr'. 
rewrite {1}Ex joinCA freeUnL; last by rewrite hdomPt inE N andbF.
rewrite Exy; split=>// z /S'. 
by rewrite {1}E !inE /= !inE /= !hdomPtUn !inE !hvalidPtUn.
Qed.

Lemma null_domE h (g : graph h) c x : dom (null_edge g c x) =i dom h.
Proof.
rewrite /null_edge => y; case: edgeP=>[b xl xr E _ _ _|//]. simpl.
by case: b E=>[|//] E; rewrite {2}E !hdomPtUn !inE !hvalidPtUn.
Qed.

Lemma null_edgeE h (g : graph h) c x (g' : graph (null_edge g c x)) y : 
        contents g' y = 
        if (mark g x) && (x == y) then 
          if c is Left then (true, null, edgr g x) 
          else (true, edgl g x, null)
        else contents g y.
Proof.
rewrite /null_edge; case: edgeP; last first.
- case: edgeP=>/=; last by case: edgeP.
  case=>xl xr E V Dx S; rewrite E in V Dx; last by case: edgeP.
  case N : (x == y)=>/=. 
  - by rewrite -(eqP N) !hdomPtUn !inE !hvalidPtUn in Dx *; rewrite Dx.
  move=>M; suff: y \notin dom h by case: edgeP.
  by rewrite E; rewrite !hdomPtUn !inE !hvalidPtUn in M *. 
move=>b xl xr; case: edgeP=>/=; last by move=>_ /edgeE ->.
case=>xl1 xr1 Ex Vx Dx _ E Vy Dy _; last by rewrite (edgeE E).
move/(hcancel2V Vy): E; case: ifP=>N [// _ E _].
by rewrite E joinCA in Ex; rewrite (edgeE Ex).
Qed.

(* A number of monotonicity lemmas for tree, max and front. *)
Lemma path_mono x (t : ptr_set) p h1 h2 (g1 : graph h1) (g2 : graph h2) : 
  dom h1 =i dom h2 -> {in dom t, forall y, contents g1 y = contents g2 y} ->
  x \in dom t -> {subset p <= dom t} -> path (edge g1) x p = path (edge g2) x p. 
Proof.
move=>D H Dx; elim: p x Dx=>[|z zs IH] //= x Dx S; rewrite IH.
- by rewrite (H _ Dx) D.
- by apply: S; rewrite inE eq_refl.
by move=>w Hw; apply: S; rewrite inE Hw orbT.
Qed.

Lemma tree_mono x t h1 h2 (g1 : graph h1) (g2 : graph h2) :
  dom h1 =i dom h2 -> {in dom t, forall y, contents g1 y = contents g2 y} ->
  tree (edge g1) x t -> tree (edge g2) x t.
Proof.
move=>D H [Dx Ht]; split=>// y Dy. 
case/(_ _ Dy): Ht=>p [[Ey Py Sy] Up]; exists p. 
split; first by rewrite -(path_mono D H Dx Sy). 
move=>p' [Ey' Py' Sy']; apply: Up; split=>//.
by rewrite (path_mono D H Dx Sy').
Qed.

Lemma max_mono t h1 h2 (g1 : graph h1) (g2 : graph h2) :
  dom h1 =i dom h2 -> {in dom t, forall y, contents g1 y = contents g2 y} ->
  maximal (edge g1) t -> maximal (edge g2) t.
Proof.
move=>D H [S Ht]; split=>// x y Dx Dy; apply: Ht (Dx) _.
by rewrite /edge !inE (H _ Dx) D in Dy *.
Qed.

(* Backward propagation of the fron property. *)
Lemma front_mono t u h1 h2 (g1 : graph h1) (g2 : graph h2) :
  dom h1 =i dom h2 -> {in dom t, forall y, contents g2 y = contents g1 y} -> 
  front (edge g2) t u -> front (edge g1) t u.
Proof.
move=>D H [S Ht]; split=>// x y Dx Dy; apply: Ht (Dx) _.
by rewrite /edge !inE (H _ Dx) D in Dy *.
Qed.


(****************************)
(* Spanning tree concurroid *)
(****************************)

Require Import autosplit finmap getter.

Module SpanResourceInvariant. 
Section SpanResourceInvariant. 
Variable sp : nat.

Notation ptr_set := (ptrmap_pcm_Encoded [encoded_set of unit]).

Definition spg := @Getter sp ptr_set heap_tp_Encoded Unit.

Lemma sp_flatsE (xs xo : ptr_set) (xj : heap) : flats (sp ->> [xs, xj, xo]) = xj.
Proof.
rewrite fsE flatsE /= !pcm_flatPt tp_flatPtE.
by case: decP=>// _; rewrite unitL unitR. 
Qed.

Notation getso s := (getso spg s).
Notation self s := (self spg s).
Notation other s := (other spg s). 
Notation joint s := (joint spg s).

Definition coh := 
  [Pred s | exists g : graph (joint s), 
            [/\ s = sp ->> [self s, joint s, other s], validH s, validSO s &
                forall x, x \in dom (self s \+ other s) = mark g x]].

(* Standar properties of the concurroid coherence predicate. *)
Lemma sp_cohE s : s \In coh -> s = sp ->> [self s, joint s, other s].
Proof. by case=>g []. Qed.

Lemma sp_cohSO s : s \In coh -> validSO s. 
Proof. by case=>g []. Qed.

Lemma sp_cohH s : s \In coh -> validH s. 
Proof. by case=>g []. Qed.

Lemma sp_cohfV s : s \In coh -> footV_eq s (mem_seq' [:: sp]).
Proof. by case=>g [E _ _ _]; rewrite E fsE. Qed.

Lemma sp_coh_weq s s' : 
        s \In coh -> s' \In coh -> weq typeCode (shr s) (shr s').
Proof. 
by case=>g [E _ _ _][g'][E' _ _ _]; rewrite E E' !fsE; apply/eqP/feapE/fmapE.
Qed.

Lemma sp_coh_sig s : s \In coh -> sp \-> (Unit : ptr_set) = [zunit loc s].
Proof. by case=>g [E _ _ _]; rewrite E fsE zunitPt. Qed.

Lemma sp_coh_zigzag s p : (s \< p) \In coh <-> (s \> p) \In coh.
Proof.
split=>C; case: C (C)=>g [E Vh Vso M] C.
- rewrite validH_zigzag // in Vh; rewrite zagzigVE // in M.
  rewrite validSO_zigzag in Vso; move: E Vh Vso; rewrite zagzigE=>->.
  by rewrite /coh InE /= !getE in g M *; exists g. 
rewrite -validSO_zigzag in Vso; rewrite -validH_zigzag // in Vh.
rewrite -zagzigVE // in M; move: E; rewrite 2!getE in g M *.
by rewrite -zagzigE /coh InE /= !getE; exists g.
Qed.

(* Instantiating the canonical structure with the coherence predicate coh. *)
Definition span_tree_CohMixin :=
  CohPredMixin sp_cohSO sp_cohH sp_cohfV sp_coh_weq sp_coh_sig sp_coh_zigzag.
Definition sp_coh := Eval hnf in CohPred coh span_tree_CohMixin.

Definition span_tree_GCMixin := @GCMixin _ sp_coh sp_cohE.
Canonical sp_getcoh := Eval hnf in GC spg sp_coh span_tree_GCMixin.

(* Additional lemmas about concurroid components. *)

Lemma sp_cohJ s : s \In sp_coh -> valid (joint s).
Proof. by move=>C; move: (cohH C); rewrite {1}(cohE C) /validH sp_flatsE. Qed. 

Lemma sp_cohG s : s \In sp_coh -> graph (joint s).
Proof. by case. Qed.

Lemma sp_markE s (g : graph (joint s)) x : 
        s \In sp_coh -> x \in dom (self s \+ other s) = mark g x. 
Proof. by case=>g' [_ _ _ ->]; congr (mark _ _); apply: proof_irrelevance. Qed.

Lemma markS s (g : graph (joint s)) x : 
        s \In sp_coh -> x \in dom (self s) -> mark g x.
Proof. by move=>C D; rewrite -sp_markE // domUn inE D (cohVSO C). Qed.

End SpanResourceInvariant. 

Module Exports.
Notation sp_coh := sp_coh.
Notation sp_flatsE := sp_flatsE.
Notation spg := spg.
Notation sp_getcoh := sp_getcoh.
Notation sp_cohJ := sp_cohJ.
Notation sp_cohG := sp_cohG.
Notation sp_markE := sp_markE.
Notation markS := markS.
Canonical sp_getcoh.

End Exports.

End SpanResourceInvariant. 

Export SpanResourceInvariant.Exports.


Module SpanTransitions.
Section SpanTransitions. 
Variable sp : nat. 

Notation ptr_set := (ptrmap_pcm_Encoded [encoded_set of unit]).
Notation coh := (sp_coh sp).
Notation getso s := (getso (spg sp) s).
Notation self s := (self (spg sp) s).
Notation other s := (other (spg sp) s).
Notation joint s := (joint (spg sp) s).

Section MarkTransition.
Variable x : ptr.

Definition mark_rel s s' := 
  exists g : graph (joint s),
    [/\ ~~ mark g x, joint s' = mark_node g x, 
        self s' = #x \+ self s, other s' = other s & s \In coh /\ s' \In coh].

Lemma mark_coh s s' : mark_rel s s' -> s \In coh /\ s' \In coh.
Proof. by case=>g []. Qed.

Lemma mark_env s s' : mark_rel s s' -> env s = env s'.
Proof. by case=>g [_ _ _ E][C C']; rewrite (cohE C) (cohE C') !fsE /= E. Qed.

Lemma mark_zagzig s s' p : 
 env s = env s' -> mark_rel (s \> p) (s' \> p) -> mark_rel (s \< p) (s' \< p).
Proof.
move=>E [g][M J S O]; case=>/coh_zigzag C /coh_zigzag C'.
rewrite /mark_rel !getE in g M J *; exists g.
by rewrite !getzE in S; rewrite !getzC //= S joinA E.
Qed.

Lemma mark_footH s s' : mark_rel s s' -> footeqH (flats s) (flats s').
Proof.
case=>g [M J S O][C C']; rewrite (cohE C) (cohE C') /= !sp_flatsE.
rewrite J /mark_node; case: edgeP=>[b xl xr E /= V _ _|//]. 
by rewrite {1}E; apply: footeqHUnL; apply: footeqHPt.
Qed.

End MarkTransition.


Section NullifyTransition.
Variables (x : ptr) (c : side).

Definition nullify_rel s s' := 
  exists g : graph (joint s), 
    [/\ x \in dom (self s), joint s' = null_edge g c x,
        self s' = self s, other s' = other s & s \In coh /\ s' \In coh].
 
Lemma nullify_coh s s' : nullify_rel s s' -> s \In coh /\ s' \In coh.
Proof. by case=> g []. Qed.

Lemma nullify_env s s' : nullify_rel s s' -> env s = env s'.
Proof. by case=>g [_ _ _ O][C C']; rewrite (cohE C) (cohE C') !fsE /= O. Qed.

Lemma nullify_zagzig s s' p : 
  env s = env s' -> nullify_rel (s \> p) (s' \> p) -> nullify_rel (s \< p) (s' \< p).
Proof.
move=>E [g][D J S _]; case=>/coh_zigzag C /coh_zigzag C'.
rewrite /nullify_rel !getE in g D J S *; exists g. 
rewrite !getzC //= S E domUn inE D andbT; split=>//.
by move: (cohVSO C); rewrite getzC //=; apply: validL. 
Qed.

Lemma nullify_footH s s' : nullify_rel s s' -> footeqH (flats s) (flats s').
Proof.
case=>g [D J S O][C C']; rewrite (cohE C) (cohE C') /= !sp_flatsE J.
rewrite /null_edge; case: edgeP=>[[|//] xl xr E /= V _ _|//].
by rewrite {1}E; apply: footeqHUnL; apply: footeqHPt.
Qed.

End NullifyTransition.

End SpanTransitions.

Module Exports. 
Section Exports. 
Variable sp : nat.

(* mark internal transition *)
Definition MarkGuarMixin x := 
  GuarMixin (@mark_coh sp x) (@mark_env sp x) (@mark_zagzig sp x). 
Canonical mark_guar x := Eval hnf in Guar (@mark_rel sp x) (MarkGuarMixin x). 
Definition markIntGuarMixin x := IntGuarMixin (@mark_footH sp x). 
Canonical mark_intguar x := 
  Eval hnf in IntGuar (mark_guar x) (markIntGuarMixin x).

(* nullify internal transition *)
Definition NullifyGuarMixin x c := 
  GuarMixin (@nullify_coh sp x c) (@nullify_env sp x c) 
            (@nullify_zagzig sp x c). 
Canonical nullify_guar x c := 
  Eval hnf in Guar (@nullify_rel sp x c) (NullifyGuarMixin x c). 
Definition nullifyIntGuarMixin x c := IntGuarMixin (@nullify_footH sp x c). 
Canonical nullify_intguar x c := 
  Eval hnf in IntGuar (nullify_guar x c) (nullifyIntGuarMixin x c).

(* the concurroid *)
Canonical SpanCore : mod_core := 
  Mod.Core [PredU [Pred r | exists x, r = mark_intguar x] & 
                  [Pred r | exists x c, r = nullify_intguar x c]]. 
Canonical SpanTree : mod := @Mod.Make SpanCore Pred0. 

End Exports.
End Exports.

End SpanTransitions.

Export SpanTransitions.Exports.


(*************************)
(* Spanning tree actions *)
(*************************)

Module SpanActions.
Section SpanActions.
Variable sp : nat. 

Notation ptr_set := (ptrmap_pcm_Encoded [encoded_set of unit]).
Notation coh := (sp_coh sp).
Notation getso s := (getso (spg sp) s).
Notation self s := (self (spg sp) s).
Notation other s := (other (spg sp) s).
Notation joint s := (joint (spg sp) s).
Notation guar := (Mod.internal (SpanTree sp)).

(* An action, trying to mark the node, and returning a boolean
   indicating a success or a failure. It is equivalen to CAS
   operationally. *)

Section TryMarkAction.
Variable x : ptr. 

Definition trymark_safe s := 
  x \in dom (joint s) /\ s \In coh. 

Definition trymark_step s s' (v : bool) := 
  exists g : graph (joint s), 
    [/\ x \in dom (joint s),
        if mark g x then [/\ v = false, joint s' = joint s & self s' = self s]
        else [/\ v = true, joint s' = mark_node g x & self s' = #x \+ self s], 
        other s' = other s &
        s \In coh /\ s' \In coh].

Lemma trymark_safe_coh : safe_coh coh trymark_safe.
Proof. by move=>s []. Qed.

Lemma trymark_step_safe : step_safe trymark_safe trymark_step.
Proof. by move=>s s' v [g][D _ _][C _]. Qed.

Lemma trymark_safe_zagzig : safe_zagzig trymark_safe.
Proof. by move=>s p [D] /coh_zigzag C; split=>//; rewrite !getE in D *. Qed.

Lemma trymark_step_guar : step_guar guar trymark_step.
Proof. 
move=>s s' v [g][]; case: edgeP=>//=; case=>xl xr J _ _ _ _ [_ E S O][C C'].
- by left; rewrite {1}(cohE C) (cohE C') /= S E O.
right; exists (mark_intguar sp x); split; first by left; exists x.
by exists g; rewrite (edgeE J). 
Qed.

Lemma trymark_step_total : step_total trymark_safe trymark_step.
Proof. 
move=>s [D C]; case: (C)=>g _; case: edgeP D=>//; case=>xl xr J V D S _.
- by exists s, false, g; rewrite (edgeE J). 
exists (sp ->> [#x \+ self s : ptr_set, mark_node g x, other s]), true, g. 
rewrite (edgeE J) /= -!toPredE /= !getE; do !split=>//.
have Vso : valid (#x \+ self s \+ other s).
- by rewrite -joinA um_validPtUn (cohVSO C) (sp_markE g) //= (edgeE J). 
exists (mark_nodeG g x); split=>//.
- by rewrite /validH sp_flatsE; apply: validG (mark_nodeG _ _). 
- by rewrite /validSO !fsE zipPt allvalidPt.
move=>y; rewrite -joinA um_domPtUn inE joinA Vso (sp_markE g) //=.
by rewrite mark_nodeE D; case: eqP.
Qed.

Lemma trymark_step_frame : step_frame trymark_safe trymark_step.
Proof. 
move=>s s' t v S [g][D X O][C C'].
case: edgeP D X=>//; case=>/= xl xr J V D Sx _ [V' J' S'].
- exists s; split; first by rewrite (cohE C) /= -S' -J' -O -(cohE C'). 
  rewrite /trymark_step !getE /= in g D O J V *.
  by exists g; rewrite (edgeE J); do !split=>//; apply/coh_zigzag. 
set s'' := sp ->> [#x \+ self s : ptr_set, mark_node g x, other s].
have E : s' = s'' \< t.
- rewrite (zigSOE C) /= (cohE C') /= S' O getE J' /s'' !fsE zigE zipPt -!fsE.
  by rewrite getzC //= joinA joinC.  
exists s''; split=>//; rewrite {}E /trymark_step {}/s'' in C' *. 
rewrite !getE in g V D J J' C' *; exists g.
by rewrite !getzC //= !getE (edgeE J); do !split=>//; apply/coh_zigzag. 
Qed.

Definition try_mark_pre_act := 
  Action trymark_safe_coh trymark_safe_zagzig trymark_step_safe 
         trymark_step_guar trymark_step_frame trymark_step_total.

Lemma trymark_step_erase : int_step_erase (try_mark_pre_act).
Proof.
move=>s1 h1 t1 v1 s2 h2 t2 v2 V E H1 H2; move: H1 H2 V E.
case=>g1 []; case: edgeP=>//= b1 xl1 xr1 J1 V1 D1 _ _ X1 O1 [C1 C1'].
case=>g2 []; case: edgeP=>//= b2 xl2 xr2 J2 V2 D2 _ _ X2 O2 [C2 C2'].
rewrite (cohE C1) (cohE C2) (cohE C1') (cohE C2') /= !sp_flatsE J1 J2 -!joinA.
move=>V /(hcancelV V) []; case=>??? {V} V E; subst b2 xl2 xr2.
case: b1 X1 X2 J1 J2; case=>->-> _ [->-> _] J1 J2.
- by rewrite J1 J2 -!joinA E.
by rewrite /mark_node D1 D2 (edgeE J1) (edgeE J2) /= -!joinA E. 
Qed.

Definition trymark_act := IntAction trymark_step_erase.

End TryMarkAction.

(* An action, removing an adge from the node x to the child, indicated
   by the side c. It's equivalent to writing the pointer
   atomically. *)

Section NullifyAction.
Variables (x : ptr) (c : side). 

Definition nullify_safe s := x \in dom (self s) /\ s \In coh. 

Definition nullify_step s s' (v : unit) := 
  exists g : graph (joint s), 
    [/\ x \in dom (self s), joint s' = null_edge g c x,
        self s' = self s, other s' = other s & s \In coh /\ s' \In coh].

Lemma nullify_safe_coh : safe_coh coh nullify_safe.
Proof. by move=>s []. Qed.

Lemma nullify_step_safe : step_safe nullify_safe nullify_step.
Proof. by move=>s s' [[g]][D J S O][]. Qed.

Lemma nullify_safe_zagzig : safe_zagzig nullify_safe.
Proof. 
move=>s p [D] /coh_zigzag C; split=>//; rewrite getE in D.
by move/validL: (cohVSO C); rewrite getzC //= domUn inE D andbT.
Qed.

Lemma nullify_step_guar : step_guar guar nullify_step.
Proof. 
move=>s s' [][g][D J S O][C C']; right; exists (nullify_intguar sp x c).
by split; [right; exists x, c | exists g].
Qed.

Lemma nullify_step_total : step_total nullify_safe nullify_step.
Proof. 
move=>s [D C]; case: (C)=>g _.
exists (sp ->> [self s, null_edge g c x, other s]), tt, g.
rewrite !getE; do !split=>//; rewrite -toPredE /= !getE.
exists (null_edgeG g c x); split=>//.
- by rewrite /validH sp_flatsE; apply: validG (null_edgeG g c x).
- by rewrite /validSO !fsE /= zipPt allvalidPt (cohVSO C).
move=>y; rewrite sp_markE // null_edgeE.
by case: ifP=>//; case: c=>/andP [M] /eqP <-.
Qed.

Lemma nullify_step_frame : step_frame nullify_safe nullify_step.
Proof. 
move=>s s' t [][D _][g][D' J S O][C C'].
set s'' := sp ->> [self s, null_edge g c x, other s].
have E : s' = s'' \< t.
- rewrite (zigSOE C) (cohE C') S J O /s'' !fsE zigE zipPt -!fsE. 
  by rewrite [other _]getE getzC //= joinC.
exists s''; split=>//; rewrite {}E /nullify_step {}/s'' in C' *. 
rewrite !getE in g D J C' *; exists g.
by rewrite !getzC //= !getE; do !split=>//; apply/coh_zigzag.
Qed.

Definition nullify_pre_act := 
  Action nullify_safe_coh nullify_safe_zagzig nullify_step_safe 
         nullify_step_guar nullify_step_frame nullify_step_total.

Lemma nullify_step_erase : int_step_erase (nullify_pre_act).
Proof.
move=>s1 h1 t1 [] s2 h2 t2 [] V E H1 H2; move: H1 H2 V E.
case=>g1 [D1] J1 S1 O1 [C1 C1'][g2][D2] J2 S2 O2 [C2 C2']. 
rewrite (cohE C1) (cohE C2) (cohE C1') (cohE C2') /= !sp_flatsE J1 J2.
rewrite /null_edge.
case: edgeP (markS g1 C1 D1)=>//; case=>//= xl1 xr1 J1' _ _ _ _.
case: edgeP (markS g2 C2 D2)=>//= b xl2 xr2 J2' _ _ _ _.
rewrite {-3}J1' {-2}J2' -!joinA => V /(hcancelV V) [] {V J1' J2'}.
by case=><-<-<- _ ->; case: c; rewrite -joinA.
Qed.

Definition nullify_act := IntAction nullify_step_erase.

End NullifyAction.


(* An action, reading the child of the node pointe x in the direction
   c. It's equivalent to reading a pointer atomically. *)


Section ReadChildAction.
Variables (x : ptr) (c : side). 

Definition read_child_safe s := x \in dom (self s) /\ s \In coh.

Definition read_child_step s s' (v : ptr) := 
  exists g : graph (joint s), 
    [/\ x \in dom (self s), 
        v = if c is Left then edgl g x else edgr g x, 
        s' = s & s \In coh].

Lemma read_child_safe_coh : safe_coh coh read_child_safe.
Proof. by move=>s []. Qed.

Lemma read_child_step_safe : step_safe read_child_safe read_child_step.
Proof. by move=>s s' v [g][]. Qed.

Lemma read_child_safe_zagzig : safe_zagzig read_child_safe.
Proof. 
move=>s p [D] /coh_zigzag C; split=>//; rewrite getE in D.
by move: (cohVSO C); rewrite getzC //= domUn inE D andbT; apply: validL.
Qed.

Lemma read_child_step_guar : step_guar guar read_child_step.
Proof. by move=>s s' v [g][_ _ -> C]; left. Qed.

Lemma read_child_step_total : step_total read_child_safe read_child_step.
Proof. 
move=>s [D C]; case: (C)=>g _.
exists s, (if c is Left then edgl g x else edgr g x), g.
by case: c.
Qed.

Lemma read_child_step_frame : step_frame read_child_safe read_child_step.
Proof. 
move=>s _ t v [D _][g][_ E -> C]; exists s; split=>//.
rewrite /read_child_step !getE in g E D *; exists g. 
by split=>//; apply/coh_zigzag.
Qed.

Definition read_child_pre_act := 
  Action read_child_safe_coh read_child_safe_zagzig read_child_step_safe 
         read_child_step_guar read_child_step_frame read_child_step_total.

Lemma read_child_step_erase : int_step_erase (read_child_pre_act).
Proof.
move=>s1 h1 t1 v2 s2 h2 t2 v3 V E H1 H2; move: H1 H2 V E.
case=>g1 [D1 ->-> C1][g2][D2 ->-> C2]. 
case: edgeP (markS g1 C1 D1)=>//; case=>//= xl1 xr1 J1 _ _ _ _.
case: edgeP (markS g2 C2 D2)=>//; case=>//= xl2 xr2 J2 _ _ _ _.
rewrite (cohE C1) (cohE C2) !sp_flatsE /= {}J1 {}J2 -!joinA.
by move=>V /(hcancelV V) []; case=><-<- _ ->. 
Qed.

Definition read_child_act := IntAction read_child_step_erase.

End ReadChildAction.


End SpanActions.

Module Exports.
Notation trymark_act := trymark_act.
Notation nullify_act := nullify_act.
Notation read_child_act := read_child_act.
End Exports.

End SpanActions.

Export SpanActions.Exports.


(******************************)
(* Stability and main program *)
(******************************)

Section SpanStabilityMain.
Variable sp : nat.
Notation W := (SpanTree sp).
Notation ptr_set := (ptrmap_pcm_Encoded [encoded_set of unit]).
Notation coh := (sp_coh sp).
Notation getso s := (getso (spg sp) s).
Notation self s := (self (spg sp) s).
Notation other s := (other (spg sp) s).
Notation joint s := (joint (spg sp) s).
Notation guar := (Mod.internal W).

Section SubgraphDef.
Variables (p : pred ptr) (s1 s2 : state).
Variables (g1 : graph (joint s1)) (g2 : graph (joint s2)).

(* Package together all the stability properties. *)

Definition subgraph :=
  [/\ dom (joint s1) =i dom (joint s2), 
      {subset dom (self s1) <= dom (self s2)} ,
      {subset dom (other s1) <= dom (other s2)} ,
      forall y, (y \in p) && (y \in dom (self s1)) ->
        contents g2 y = contents g1 y,
      forall y, ~~ mark g2 y -> contents g1 y = contents g2 y & 
      forall x, (edgl g2 x \in [:: null; edgl g1 x]) &&
                (edgr g2 x \in [:: null; edgr g1 x])].

(* names for the projections *)
Definition subgrD (x : subgraph) := let: (And6 y _ _ _ _ _) := x in y.
Definition subgrS (x : subgraph) := let: (And6 _ y _ _ _ _) := x in y.
Definition subgrO (x : subgraph) := let: (And6 _ _ y _ _ _) := x in y.
Definition subgrM (x : subgraph) := let: (And6 _ _ _ y _ _) := x in y.
Definition subgrN (x : subgraph) := let: (And6 _ _ _ _ y _) := x in y.
Definition subgrE (x : subgraph) := let: (And6 _ _ _ _ _ y) := x in y.

End SubgraphDef.

Notation subgraphT g1 g2 := (subgraph xpredT g1 g2).

Lemma subgr_refl p s g : @subgraph p s s g g.
Proof. by split=>// z; rewrite !inE !eq_refl !orbT. Qed.

Hint Resolve subgr_refl.

Lemma subgr_trans p1 p2 p3 s1 s2 s3 g1 g2 g3 : 
       [predI p1 & p2] =i p3 -> @subgraph p1 s1 s2 g1 g2 -> 
       @subgraph p2 s2 s3 g2 g3 -> @subgraph p3 s1 s3 g1 g3.
Proof.
move=>E [D1 S1 O1 M1 N1 Ed1][D2 S2 O2 M2 N2 Ed2]; split.
- by move=>z; rewrite D1 D2.
- by move=>z /S1 /S2.
- by move=>z /O1 /O2.
- move=>z /andP /=; rewrite -E; case=>/andP [/= X1 X2] Y. 
  by rewrite -M1 ?X1 // -M2 // X2; apply: S1.
- move=>y My; move: (My); rewrite -N2 //.
  by move=>My2; apply: N1 My2; rewrite D2.
move=>y; case/andP: (Ed1 y)=>X1 X2; case/andP: (Ed2 y)=>Y1 Y2.
rewrite !inE in X1 X2 Y1 Y2 *.
by case/orP: Y1=>/eqP ->; [|rewrite X1]; case/orP: Y2=>/eqP ->.
Qed.

Lemma subgr_transT s1 s2 s3 g1 g2 g3 : 
        @subgraph xpredT s1 s2 g1 g2 -> 
        @subgraph xpredT s2 s3 g2 g3 -> @subgraph xpredT s1 s3 g1 g3.
Proof. by apply: subgr_trans. Qed.

Lemma subgrX s1 s2 g1 g2 : 
        @subgraph xpredT s1 s2 g1 g2 <-> subgraph (dom (self s1)) g1 g2. 
Proof. 
by split; case=>??? M ?; split=>// y X; apply: M; rewrite andbb in X *. 
Qed.

Lemma subgr_leq p1 p2 s1 s2 g1 g2 : 
        {subset p1 <= p2} -> 
        @subgraph p2 s1 s2 g1 g2 -> @subgraph p1 s1 s2 g1 g2.
Proof. by move=>P [??? M ?]; split=>// y X; apply: M; case/andP: X=>/P ->. Qed.

Lemma subgr_mrely s1 s2 g1 g2 : mrely W s1 s2 -> @subgraph xpredT s1 s2 g1 g2. 
Proof.
move=>M; case: (mrely_coh M)=>C1 C2; case: M; last first.
- by case=>ch [h][/mem_chan][ct][].
rewrite /subgraph=>G; case: G g2.
- case=>/transp_inj <- _ gx; rewrite (proof_irrelevance g1 gx). 
  by split=>// x; rewrite !inE !eq_refl !orbT.
case=>G [[]].
- case=>y -> [/= g'][M -> O ->] _ gx; rewrite (proof_irrelevance g1 g').
  split=>z; rewrite ?mark_domE ?mark_nodeE //.
  - by move=>D; rewrite O um_domPtUn inE D -O (validR (cohVSO C2)) orbT. 
  - move=>D; case: edgeP (M)=>//; case=>//= xl xr *. 
    by case: ifP (markS g' C1 D) M=>// /eqP <- ->. 
  - by case: edgeP (M)=>//; case=>// xl xr; case: ifP.
  case: edgeP (M)=>/=; last by rewrite !inE !eq_refl !orbT.
  case=>// xl xr J *; case: ifP; last by rewrite !inE !eq_refl !orbT.
  by move=>/eqP <-; rewrite (edgeE J) /= !inE !eq_refl !orbT.
case=>y [c ->][/= g'][Dy ->->->] _ gx; rewrite (proof_irrelevance g1 g').
split=>z; rewrite ?null_domE ?null_edgeE //.
- move=>Dx; case: validUn (cohVSO C1)=>//= _ _ /(_ _ Dx) {Dx}.
  by case: ifP Dy=>// /andP [_ /eqP ->] ->.
- by case: ifP=>//; case: c {gx}.
case: ifP Dy; last by rewrite !inE !eq_refl !orbT.
by case/andP=>M /eqP ->; case: c {gx}; rewrite !inE !eq_refl /= orbT.
Qed.

Lemma subgr_steps s1 s2 g1 g2 : 
        menv_steps W s1 s2 -> @subgraph xpredT s1 s2 g1 g2.
Proof.
case=>n; elim: n s1 g1=>[|n IH] /= s1 g1.
- by case=>E; rewrite E in g1 *; rewrite (proof_irrelevance g1 g2).
case=>s [R S]; move/proj2: (mrely_coh R)=>C.
by apply: subgr_transT (IH _ (sp_cohG C) S); apply: subgr_mrely.
Qed.

Lemma dom_steps s1 s2 : menv_steps W s1 s2 -> dom (joint s1) =i dom (joint s2).
Proof.
move=>M; case: (menvs_coh M)=>C1 C2.
by case/(subgr_steps (sp_cohG C1) (sp_cohG C2)): M.
Qed.

Lemma other_steps s1 s2 : 
        menv_steps W s1 s2 -> {subset dom (other s1) <= dom (other s2)}.
Proof. 
move=>M; case: (menvs_coh M)=>C1 C2.
by case/(subgr_steps (sp_cohG C1) (sp_cohG C2)): M.
Qed.

Lemma marked_steps s1 s2 (g1 : graph (joint s1)) (g2 : graph (joint s2)) x : 
        menv_steps W s1 s2 -> x \in dom (self s1) -> 
        contents g2 x = contents g1 x.
Proof. by move/(subgr_steps g1 g2)/subgrM; apply. Qed.

Lemma unmarked_steps s1 s2 (g1 : graph (joint s1)) (g2 : graph (joint s2)) x : 
        menv_steps W s1 s2 -> ~~ mark g2 x -> contents g1 x = contents g2 x. 
Proof. by move/(subgr_steps g1 g2)/subgrN; apply. Qed.

Lemma subgr_marked (p : pred _) s1 s2 g1 g2 x : 
        @subgraph p s1 s2 g1 g2 -> mark g1 x -> mark g2 x.
Proof. by move=>S; apply: contraLR=>M; rewrite (subgrN S M). Qed.

Lemma mark_steps s1 s2 (g1 : graph (joint s1)) (g2 : graph (joint s2)) x : 
        menv_steps W s1 s2 -> mark g1 x -> mark g2 x.
Proof. by move/(subgr_steps g1 g2); apply: subgr_marked. Qed.

Lemma edge_steps s1 s2 (g1 : graph (joint s1)) (g2 : graph (joint s2)) x : 
        menv_steps W s1 s2 ->  
        (edgl g2 x \in [:: null; edgl g1 x]) &&
        (edgr g2 x \in [:: null; edgr g1 x]).
Proof. by move/(subgr_steps g1 g2)/subgrE; apply. Qed.

(* helper equalities to supply in subg_trans *)
Lemma meetTp T (p : pred T) : [predI predT & p] =i p.
Proof. by []. Qed.

Lemma meetpT T (p : pred T) : [predI p & predT] =i p.
Proof. by move=>z; rewrite inE andbT. Qed.

Lemma meetpp T (p : pred T) : [predI p & p] =i p.
Proof. by move=>z; rewrite inE andbb. Qed.

(*************************)
(* Action specifications *)
(*************************)

Program Definition trymark x :
  {s1 (g1 : graph (joint s1))}, STsep [W] 
    (fun i => s1 = i /\ x \in dom (joint i),
     fun b s2 => exists (g2 : graph (joint s2)),
       [/\ subgraphT g1 g2, 
           self s2 = if b then #x \+ self s1 else self s1,
           mark g2 x = true &
           b -> contents g1 x = (false, edgl g2 x, edgr g2 x)]) :=
  Do (act (trymark_act sp x)).
Next Obligation.
apply: gh=>_ [s1 g1][<- Dx] C1; apply: val_act=>i Mi.
rewrite (menvs_loc Mi); rewrite (dom_steps Mi) in Dx. 
split; first by case: (menvs_coh Mi).
move=>y k s2 [gi][_ X O][Ci _] Mk; case: (menvs_coh Mk)=>Ck C2.
have [gk g2] : graph (joint k) /\ graph (joint s2) by split; apply: sp_cohG.
move: (subgr_steps g1 gi Mi) (subgr_steps gk g2 Mk)=>Si Sk.
rewrite -(menvs_loc Mk); exists g2; case: ifP X=>Mx [-> J S].
- split=>//; last by move/mark_steps: Mk; apply; rewrite (eqG _ _ _ J).
  apply: subgr_transT Si (subgr_transT _ Sk).
  by split=>z *; rewrite ?J ?S ?O ?(eqG _ _ _ J) // !inE !eq_refl !orbT.
split=>//; last first.
- rewrite (unmarked_steps g1 Mi (negbT Mx)) {Sk}.
  rewrite (marked_steps gk g2 Mk); last first.
  - by move: (cohVSO Ck); rewrite S um_domPtUn inE eq_refl => /validL ->. 
  by rewrite J in gk *; rewrite mark_nodeE /= Dx eq_refl /= -Mx; case: edgeP.
- apply: {Sk} (mark_steps g2 Mk).
  by rewrite J in gk *; rewrite mark_nodeE Dx eq_refl. 
apply: subgr_transT Si (subgr_transT _ Sk); split.
- by move=>z; rewrite J mark_domE.
- by move=>z; move: (cohVSO Ck); rewrite S domUn inE orbC => /validL ->->.
- by move=>z; rewrite O.
- move=>z Dz {Sk}; rewrite J in gk *.
  by rewrite mark_nodeE Dx /=; case: eqP Mx=>// ->; rewrite (markS gi Ci Dz).
- move=>z Mz {Sk}; rewrite J in gk Mz *; rewrite mark_nodeE Dx /=.
  by case: eqP Mz=>// <-; rewrite mark_nodeE Dx eq_refl.
move=>z {Sk}; rewrite J in gk *; rewrite mark_nodeE Dx /=. 
by case: eqP=>[->|]; rewrite !inE !eq_refl !orbT.
Qed.

Program Definition read_child x c :
  {s1 (g1 : graph (joint s1))}, STsep [W] 
    (fun i => s1 = i /\ x \in dom (self i),
     fun r s2 => exists (g2 : graph (joint s2)),
       [/\ subgraphT g1 g2, self s2 = self s1 &
           r = if c is Left then edgl g1 x else edgr g1 x]) :=
  Do (act (read_child_act sp x c)).
Next Obligation.
apply: gh=>_ [s1 g1][<- Dx] C1; apply: val_act=>i Mi.
rewrite (menvs_loc Mi) in Dx *; split; first by case: (menvs_coh Mi). 
move=>y _ s2 [gi][_ ->->] Ci Mk; case: (menvs_coh Mk)=>_ C2.
have g2 : graph (joint s2) by apply: sp_cohG.
rewrite -(menvs_loc Mk); exists g2; split=>//.
- by apply: subgr_trans (subgr_steps g1 gi Mi) (subgr_steps gi g2 Mk). 
by rewrite (marked_steps g1 gi Mi) ?(menvs_loc Mi) //; case: c.
Qed.

Program Definition nullify x c :
  {s1 (g1 : graph (joint s1)) t}, STsep [W] 
    (fun i => s1 = i /\ self i = #x \+ t, 
     fun (_ : unit) s2 => exists g2 : graph (joint s2),
        [/\ subgraph (dom t) g1 g2, self s2 = self s1 & 
            contents g2 x = if c is Left then (true, null, edgr g1 x) 
                            else (true, edgl g1 x, null)]) := 
  Do (act (nullify_act sp x c)).
Next Obligation. 
apply: gh=>_ [s1][/= g1 t][<-] S C1; apply: val_act=>i Mi. 
have Vx : valid (self s1) by apply: validL (cohVSO C1).
have Dx : x \in dom (self s1) by rewrite S um_domPtUn inE eq_refl -S Vx. 
rewrite (menvs_loc Mi)in Vx Dx S *; split; first by case: (menvs_coh Mi).
case=>k s2 [gi][_ Jk Sk Ok][Ci Ck] Mk; case: (menvs_coh Mk)=>_ C2.
have [gk g2] : graph (joint k) /\ graph (joint s2) by split; apply: sp_cohG.
rewrite -(menvs_loc Mk); exists g2; split=>//; last first.
- rewrite -(marked_steps _ _ Mi) ?(menvs_loc Mi) // (marked_steps _ _ Mk) ?Sk //.
  rewrite Jk in gk *; rewrite null_edgeE eq_refl (markS gi Ci Dx).
  by case: c {Jk gk}.
apply: subgr_trans (meetpT _) _ (subgr_steps gk g2 Mk).
apply: subgr_trans (meetTp _) (subgr_steps g1 gi Mi) _.
split.
- by move=>z; rewrite Jk null_domE.
- by rewrite Sk.
- by move=>z; rewrite -Ok.
- move=>y /andP [Dy _]; move: Vx; rewrite Jk in gk *.
  rewrite null_edgeE (markS gi Ci Dx) /=.
  by rewrite S um_validPtUn andbC; case: ifP Dy=>// /eqP -> ->.
- move=>y; rewrite Jk in gk *; rewrite null_edgeE (markS gi Ci Dx) /=. 
  by case: ifP; case: c {Jk gk}=>//= _; apply: unmarked_steps Mi.
move=>y; rewrite Jk in gk *; rewrite null_edgeE.
case: ifP; last by rewrite !inE !eq_refl !orbT.
by case/andP=>_ /eqP ->; case: c {Jk gk}=>/=; rewrite !inE !eq_refl !orbT.
Qed.

(********************************************************)
(* Specification and verification of the span procedure *)
(********************************************************)

Notation span_post x s1 g1 := 
  (fun r s2 => exists g2 : graph (joint s2), subgraphT g1 g2 /\ 
     if r then exists t,
       [/\ self s2 = self s1 \+ t, x != null, 
           tree (edge g2) x t, maximal (edge g2) t & 
           front (edge g1) t (self s2 \+ other s2)]
     else [/\ self s2 = self s1 & (x == null) || mark g2 x]). 

Definition span_tp := forall x,
  {i (g1 : graph (joint i))}, STsep [W] 
    (fun s1 => i = s1 /\ (x == null) || (x \in dom (joint s1)), span_post x i g1). 

Program Definition span : span_tp :=
  ffix (fun (loop : span_tp) x =>
    Do (if x == null then ret false else 
        r <-- trymark x;
        if r then
          xl <-- read_child x Left;
          xr <-- read_child x Right;
          rs <-- par (loop xl) (loop xr);
          (if rs.1 then ret tt else nullify x Left);;
          (if rs.2 then ret tt else nullify x Right);;
          ret true
        else ret false)).
Next Obligation.

(* "Pull out" the logical variables and consider two pbranches of the
   first if-statement depending on whether x is null or not. *)

apply: gh=>_ [s1 g1][<- Dx] C1; case: ifP Dx=>/= [/eqP -> _|_ Dx].

(* Apply the rule val_ret for proving the specification of the (ret
   false) statement in the given context. Demonstrate the exact graph
   g2 as a witness for the existential quantification. *)

- apply: val_ret=>s2 M; case: (menvs_coh M)=>_ /sp_cohG g2; exists g2. 

(* Prove the remaining conjunction by using two stability lemmas:
   subgr_steps and menvs_loc. *)

  by split; [apply: subgr_steps M | rewrite (menvs_loc M)].

(* Apply the step lemm, implementing the rule for sequential
   composition. Then, instantiate trymark's logical variables s1 and
   g1 apropriately and apply the rule val_do for external function
   invocation to use the specification of trymark, proved earlier. *)

apply: step; apply: (gh_ex s1); apply: (gh_ex g1); apply: val_do=>//.

(* Case-analyse on x0 being true and false, swap proof goals. *)

case; last first.

(* The case of 'false', prove by applying the val_ret rule and using a
   number of stability-related fact. *)

- move=>i1 [gi1][Sgi Si Mxi _] Ci1. 
  apply: val_ret=>i2 M; case: (menvs_coh M)=>_ /sp_cohG g2; exists g2.
  split; first by apply: subgr_trans Sgi (subgr_steps _ _ M).
  by rewrite -(menvs_loc M) (mark_steps g2 M Mxi).

(* More interesting case: trymark returned true. Introduce a number of
produced facts into the context. *)

move=>i1 [gi1][Sgi Si Mxi /(_ (erefl _)) Cti] Ci1.

(* An example of forward-reasoning: we will need the fact Dxi sometime
   in the future, so let's prove it now. *)

have Dxi : x \in dom (self i1).
- by move/validL: (cohVSO Ci1); rewrite Si um_domPtUn inE eq_refl => ->.

(* Progress one step further, apply val_do rule to use the spec of
   read_child procedure; introduce a number of facts into the context. *)

apply: step; apply: (gh_ex i1); apply: (gh_ex gi1); apply: val_do=>//.
move=>_ i2 [gi2][Sgi2 Si2 ->] Ci2.

(* Keep going the same way/ *)

apply: step; apply: (gh_ex i2); apply: (gh_ex gi2); apply: val_do.

(* Prove that the precondition of the second read_child is satisfied. *)

- by rewrite Si2.

(* Perform some bookkeeping of known facts and proceed one step further. *)

move=>_ i3 [gi3][/(subgr_transT Sgi2) Sgi3 Si3 ->] Ci3.
rewrite (subgrM Sgi2 Dxi); rewrite {Sgi2 gi2 i2 Ci2}Si2 in Si3 *.
apply: step.

(* In order to apply the parallel composition rule par_do, we
   construct the "repartitioned" states, which will be given to the
   left and the right forked thread to run. *)

have Spl : sself [:: sp_getcoh sp] i3 = self i3 \+ Unit by rewrite unitR.
set i3r := sp ->> [Unit, joint i3, self i3 \+ other i3].
have gi3r : graph (joint i3r) by rewrite getE.

(* Invoke the par_do rule specifying which state goes to which thread;
   prove that the proconditions for both forked threads hold. *)

apply: (par_do (r1:=span_post (edgl gi1 x) i3 gi3) 
               (r2:=span_post (edgr gi1 x) i3r gi3r) _ Spl)=>//=.
- apply: (gh_ex i3); apply: (gh_ex gi3); apply: val_do=>//.
  - rewrite unitL -(cohE Ci3) -(subgrD Sgi3); split=>//. 
    by apply: (@edgeG _ _ x); rewrite inE eq_refl. 
- apply: (gh_ex i3r); apply: (gh_ex gi3r); apply: val_do=>// Ci3r.
  rewrite getE -(subgrD Sgi3); split=>//.
  by apply: (@edgeG _ _ x); rewrite !inE eq_refl orbT.

(* At the end of parallel composition, move the new facts, arising
  from the forked threads' postconditions, into the context and
  perform some simplifications. *)

case=>{Spl} [rl rr] i4 gsl gsr Ci4 _ _ Si' [gi4][Sg X1][gi4'][Sg'] /=; move: X1.
rewrite /subgraph !getE in gi4 gi4' Sg Sg' *. 
rewrite {}/i3r !getE in gi3r Sg' *. 
rewrite -{gi3r}(proof_irrelevance gi3 gi3r) in Sg' *.
rewrite -{gi4'}(proof_irrelevance gi4 gi4') in Sg' *.
rewrite -(subgrM Sgi3 Dxi) in Mxi Cti *; rewrite -{}Si3 in Si Dxi.
move: (subgr_transT Sgi Sgi3)=>{Sgi3 i1 gi1 Ci1 Sgi} Sgi.

(* Prove in advance a number of stability-related facts relating the
   components of the state after the parallel composition and the
   state before it. *)

have Fxr tr u : {subset dom tr <= dom gsr} -> 
  front (edge gi3) tr u -> front (edge g1) tr u.
- move=>S; apply: front_mono; first by move=>z; rewrite (subgrD Sgi).
  move=>y /S Dsr; rewrite (subgrN Sgi) // -(sp_markE gi3 y Ci3).
  apply/negP; case: Sg'=>_ _ S' _ _ _ /S'.
  move: (cohVSO Ci4); rewrite Si' -joinA joinCA.
  by case: validUn=>// _ _ /(_ _ Dsr) /negbTE ->. 
have Fxl tl u : valid (#x \+ self s1 \+ tl) -> {subset dom tl <= dom gsl} ->
  front (edge gi3) tl u -> front (edge g1) tl u.
- move=>V S; apply: front_mono; first by move=>z; rewrite (subgrD Sgi).
  move=>y Dy; rewrite /= (subgrN Sgi) // -(sp_markE gi3 y Ci3) Si.
  rewrite domUn inE -Si (cohVSO Ci3) /= negb_or Si. 
  rewrite joinC in V; case: validUn V=>// _ _ /(_ _ Dy) -> _.  
  apply/negP; case: Sg=>_ _ O _ _ _ /O. 
  move: (cohVSO Ci4); rewrite Si' -joinA. 
  by case: validUn (S _ Dy)=>// _ _ N /N /negbTE ->.
have {Sg Sg'} Sgi' : subgraphT gi3 gi4. 
- case: Sg Sg'=>D S O M N Ed [_ S' O' _ _ _]; split=>//.
  - by move=>z /S X; rewrite Si' domUn inE -Si' (validL (cohVSO Ci4)) X.
  move=>z Dz; have: z \in dom (self i3 \+ other i3).
  - by rewrite domUn inE (cohVSO Ci3) Dz orbT.
  move/(O' z); rewrite domUn inE; case/andP=>_ /orP [|//].
  move/(O z): Dz; rewrite domUn inE; case/andP=>_ /orP [L R|//].
  move: (validL (cohVSO Ci4)); rewrite Si'.
  by case: validUn L=>//_ _ /(_ _ R) /negbTE ->.
case: (Sgi')=>_ S _ E _ _; rewrite -{}E // in Mxi Cti *.
move/S: Dxi=>{S} Dxi /=; rewrite {}Si.
move: (subgr_transT Sgi Sgi')=>{Sgi Sgi'} Sgi.

(* The next part of the proof essentially considerss FOUR different
   configurations of rl and rr, elaborating on the outcomes of the
   children threads. The large part of the proof is concerned with
   preseving graph-related properties through the execution, so we
   elaborate only on the essential moments. *)


(* Consider both values of rl, starting from rl = false. *)

case: rl; last first. 
- case=>Sl Ml X; rewrite {Fxl gsl}Sl -joinA in Si' X *.
  apply: step; apply: (gh_ex i4); apply: (gh_ex gi4).
  apply: (gh_ex (self s1 \+ gsr)).

(* Now we nullify the left edge (since rl-false). *)

  apply: val_do=>//; case=>i5 [gi5][Sgi5 Si5 Cti5] Ci5. 
  rewrite -Si5 in Si' Dxi.

(*  Case-analyse on rr, starting from rr = false. *)

  case: rr X; last first.
  - case=>Sr Mr; rewrite {gsr}Sr unitR in Si' Fxr Sgi5. 
    apply: step; apply: (gh_ex i5); apply: (gh_ex gi5).

(* Since rr = false, we need to nullify the right edge as well. *)

    apply: (gh_ex (self s1)); apply: val_do=>//; case=>i6 [gi6][Sgi6 Si6].
    rewrite {}Cti5 => /= Cti' Ci6.
    move/(subgr_trans (meetpp _) Sgi5): Sgi6=>{Sgi5} Sgi6.
    rewrite -{}Si6 {gi5 Ci5} in Si' Si5 Dxi. 
    apply: val_ret=>i7 M; case: (menvs_coh M)=>_ Ci7; move: (sp_cohG Ci7)=>gi7.
    move: (subgr_trans (meetpT _) Sgi6 (subgr_steps _ gi7 M))=>{Sgi6} Sgi7.
    rewrite -(marked_steps gi6 gi7 M Dxi) in Cti'.
    rewrite (menvs_loc M) in Si5 Si' Dxi.

(* At this point it should be clear that the initial thread marked
   only x (see hypothesis Si' in the context, so we need to prove that
   the t = #x is the valid one for the span's postcondition. *)

    exists gi7; split=>//.
    - by apply/subgrX; apply: subgr_trans Sgi Sgi7. 
    exists (#x); rewrite joinC. (* Our t = #x. *)
    have X : edge gi7 x =1 pred0.
    - by move=>z; rewrite inE Cti' inE andbC; case: eqP.
    split=>//; first by [apply: tree0]; first by apply: max0.
    apply: frontPt; last by rewrite domUn inE (cohVSO Ci7) Dxi.
    move=>z; rewrite inE Cti inE; case/and3P=>_ Nz D.
    rewrite (sp_markE _ _ Ci7); apply: subgr_marked Sgi7 _.
    by case/orP: D Nz Ml Mr => /eqP -> /negbTE ->.

(* Now consider the case rr = true, rl = false *)

  case=>tr [Sr Nr Tr Mr Fr]; rewrite {gsr}Sr unitL in Fxr Fr Sgi5 Si' *. 
  rewrite joinCA joinA -(joinA (#x)) -Si' Si5 in Fr.
  move/Fxr: Fr => /(_ (fun x k => k)) {i3 gi3 Ci3 Fxr} Fr. 
  apply: step; apply: val_ret=>i6 M; apply: val_ret=>i7 /(menvs_trans M)=>{M} M.
  case: (menvs_coh M)=>_ Ci7; move: (sp_cohG Ci7)=>gi7. 
  rewrite -(marked_steps gi5 gi7 M Dxi) in Cti5.
  rewrite (menvs_loc M) in Dxi Si' Si5. 
  move/validL: (cohVSO Ci7)=>/= V; rewrite Si' in V.
  move: (subgr_trans (meetpT _) Sgi5 (subgr_steps _ gi7 M))=>{Sgi5} Sgi5.
  exists gi7; split=>{i5 gi5 Ci5 M}.
  - apply/subgrX; move/subgrX: Sgi Sgi5; apply: subgr_trans.
    by move=>z; rewrite inE /= domUn inE (validR V) orbC orKb.
  exists (#x \+ tr); rewrite joinCA; move: (subgrD Sgi5) => Di.
  have Ci : {in dom tr, forall y : ptr, contents gi4 y = contents gi7 y}.
  - move=>z Dz /=; rewrite (subgrM Sgi5) // -Si5 Si' !domUn !inE.  
    by rewrite domUn inE Dz V (validR V) /= !orbT. 
  have E: edge gi7 x =1 pred1 (edgr gi4 x).
  - move=>z /=; rewrite Cti5 inE -Di -(subgrD Sgi).
    by rewrite Dx !(eq_sym z); case: eqP=>//= <-; case: eqP Nr.
  split=>//.

(* The validity of the resulting tree is established using the tree1
   and max1 lemmas, proved at the beginning of the file. *)

  - by apply: tree1 E (tree_mono Di Ci Tr) (max_mono Di Ci Mr).
  - by apply: max1 E (proj1 Tr) (max_mono Di Ci Mr).
  apply: frontUn; last first.
  - apply: front_leq Fr=>z; rewrite !domUn !inE (cohVSO Ci7) /= Si5.
    by case/andP=>_  /orP [->//|] /(subgrO Sgi5) ->; rewrite orbT.
  apply: frontPt; last by rewrite domUn inE (cohVSO Ci7) Dxi.
  move=>z; rewrite inE Cti inE; case/and3P=>_ Nz D.
  case/orP: D Nz Ml Nr=>/eqP -> /negbTE -> /=.
  - by move/(subgr_marked Sgi5); rewrite (sp_markE _ _ Ci7).
  rewrite domUn inE (cohVSO Ci7) Si' joinA domUn inE -joinA V.
  by rewrite (proj1 Tr) orbT.

(* Now we consider the case rl = true and prove some more hypotheses,
   useful in the future (V and S). *)

case=>tl [Sl Nl Tl Ml Fl]; rewrite {gsl}Sl in Si' Fl Fxl *.
have V : valid (#x \+ self s1 \+ tl \+ gsr). 
- by move/validL: (cohVSO Ci4); rewrite Si'. 
have S: {subset dom tl <= dom (#x \+ self s1 \+ tl)}.
- by move=>z; rewrite domUn inE (validL V) orbC => ->.
move/(Fxl _ _ (validL V) S): Fl=>{Fxl} Fl X.

(* Since rl=true, we just proceed further without removing edges. *)

apply: step; apply: val_ret=>i5 M.
case: (menvs_coh M)=>_ Ci5; move: (sp_cohG Ci5)=>gi5. 
rewrite -!(joinA (#x)) in Si' V Fl. 
have Si5: self i4 = self i5 by rewrite (menvs_loc M).
move: (Dxi)=>Dxi'; rewrite Si5 in Si' Dxi'.
move: (subgr_steps gi4 gi5 M)=>{M} Sgi5.

(* Case-analysis on rr, starting from rr = false. THis case is
   symmetric to what was already considered. *)

case: rr X; last first.
- case=>Sr Mr; rewrite {gsr Fxr}Sr unitL unitR in V Si' Si5 Fl.
  apply: step; apply: (gh_ex i5); apply: (gh_ex gi5).
  apply: (gh_ex (self s1 \+ tl)); apply: val_do=>//. 
  case=>i6 [gi6][Sgi6 Si6 Cti6] Ci6. 
  rewrite (subgrM Sgi5) // in Cti6; rewrite -{}Si6 in Si' Si5 Dxi'. 
  move/(subgr_trans (meetTp _) Sgi5): Sgi6=>{Sgi5 i5 gi5 Ci5} Sgi5.
  apply: val_ret=>i7 M; case: (menvs_coh M)=>_ Ci7; move: (sp_cohG Ci7)=>gi7.
  rewrite -(marked_steps gi6 gi7 M Dxi') in Cti6.
  rewrite (menvs_loc M) in Si' Si5 Dxi'.
  move: (subgr_trans (meetpT _) Sgi5 (subgr_steps _ gi7 M))=>{Sgi5} Sgi5.
  exists gi7; split.
  - apply/subgrX; move/subgrX: Sgi Sgi5; apply: subgr_trans. 
    by move=>z; rewrite inE /= domUn inE (validR V) /= orbC orKb. 
  exists (#x \+ tl); rewrite joinCA; move: (subgrD Sgi5)=>Di.
  have Ci : {in dom tl, forall y, contents gi4 y = contents gi7 y}.
  - move=>z Dz; rewrite /= (subgrM Sgi5) // Si5 Si' !domUn !inE.  
    by rewrite domUn inE Dz V (validR V) /= !orbT. 
  have E : edge gi7 x =i pred1 (edgl gi4 x).
  - move=>z; rewrite /= inE /= -Di Cti6 inE -(subgrD Sgi) Dx /= inE.
    by rewrite !(eq_sym z) orbC; case: eqP=>//= <-; case: eqP Nl.
  split=>//.
  - by apply: tree1 E (tree_mono Di Ci Tl) (max_mono Di Ci Ml).
  - by apply: max1 E (proj1 Tl) (max_mono Di Ci Ml).
  apply: frontUn; last first.
  - apply: front_leq Fl=>z; rewrite joinA !domUn !inE (cohVSO Ci7) /= -Si'. 
    by case/andP=>_  /orP [->//|] /(subgrO Sgi5) ->; rewrite orbT.
  apply: frontPt; last by rewrite domUn inE (cohVSO Ci7) Dxi'.
  move=>z; rewrite inE Cti inE; case/and3P=>_ Nz D.
  case/orP: D Nz Mr Nl=>/eqP -> /negbTE -> /=; last first.
  - by move/(subgr_marked Sgi5); rewrite (sp_markE _ _ Ci7).
  rewrite domUn inE (cohVSO Ci7) Si' joinA domUn inE -joinA V.
  by rewrite (proj1 Tl) orbT.

(* Final case: both children threads marked something (tl and tr,
   correspondingly), *)

case=>tr [Sr Nr Tr Mr Fr]; rewrite {gsr}Sr unitL in V Fl Fxr Si' Fr.
move/Fxr: Fr=>/(_ (fun x k => k)) {Fxr} Fr.
rewrite -(joinA _ tl) in Si' V.
rewrite (joinA (_ \+ tl)) joinA -(joinA _ tl) in Fl.
rewrite joinCA joinA -(joinA _ tl) -(joinA _ (self _)) in Fr. 
have W : valid (tl \+ tr).
- by move: (cohVSO Ci5); rewrite Si'; move/validL/validR/validR.
apply: step; apply: val_ret=>i6 M.
apply: val_ret=>i7 /(menvs_trans M)=>{i6 M} M.
case: (menvs_coh M)=>_ Ci7; move: (sp_cohG Ci7)=>gi7.
move: (subgr_transT Sgi5 (subgr_steps _ gi7 M))=>{Sgi5} Sgi5.
rewrite (menvs_loc M) {i5 gi5 Ci5 M} in Si5 Si' Dxi'. 
exists gi7; split.
- by apply/subgrX; apply/subgrX; apply: subgr_trans Sgi Sgi5.

(* What remains to prove is that #x \+ (tl \+ tr) is a valid tree,
   marked by self. *)

exists (#x \+ (tl \+ tr)); rewrite joinCA; move: (subgrD Sgi5)=>Di.
have [Cil Cir] : {in dom tl, forall y, contents gi4 y = contents gi7 y} /\
                 {in dom tr, forall y, contents gi4 y = contents gi7 y}.
- split=>z Dz /=; rewrite (subgrM Sgi5) //= Si5;
  move/validL: (cohVSO Ci7); rewrite Si' (joinA (#x)) joinC;
  by rewrite domUn inE (domUn tl) inE W Dz => -> //=; rewrite orbT.
have E: edge gi7 x =i pred2 (edgl gi4 x) (edgr gi4 x).
- move=>z /=; rewrite inE /= -Di (subgrM Sgi5) //. 
  case: edgeP Nl Nr=>//= _ xl xr _ _ _ _ /negbTE Nl /negbTE Nr.
  by rewrite inE !(eq_sym z); case: eqP=>// <-; rewrite Nl Nr.
split=>//.
- by apply: tree2 E (tree_mono Di Cil Tl) (max_mono Di Cil Ml)
                    (tree_mono Di Cir Tr) (max_mono Di Cir Mr) W.
- by apply: max2 E (proj1 Tl) (max_mono Di Cil Ml) 
                   (proj1 Tr) (max_mono Di Cir Mr).
apply: frontUn; last first.
- apply: frontUn; [apply: front_leq Fl | apply: front_leq Fr]=>z;
  rewrite -Si' !domUn !inE (cohVSO Ci7); 
  by case/andP=>_  /orP [->//|] /(subgrO Sgi5) ->; rewrite orbT.
apply: frontPt; last by rewrite domUn inE (cohVSO Ci7) Dxi'.
move=>z; rewrite inE Cti inE; case/and3P=>_ _ X. 
move: (cohVSO Ci7); rewrite Si' (joinA (#x)) -(joinC (tl \+ tr)).
rewrite -(joinA (tl \+ tr)) domUn inE domUn inE W => -> /=. 
by case/orP: X=>/eqP ->; rewrite ?(proj1 Tl) ?(proj1 Tr) ?orbT.
Qed.

End SpanStabilityMain.

Module SpanDecorate.
Section SpanDecorate.
Variable sp : nat.
Notation W := (SpanTree sp).
Notation ptr_set := (ptrmap_pcm_Encoded [encoded_set of unit]).
Notation coh := (sp_coh sp).
Notation self s := (self (spg sp) s).
Notation other s := (other (spg sp) s).
Notation joint s := (joint (spg sp) s).

Definition graph_dec (g : ptr_set * heap) s := 
  exists (pf : graph g.2),
    [/\ s = sp ->> [g.1, g.2, Unit] & s \In coh].


Lemma dec_prec g h1 h2 s1 s2 :
        valid (flats s1 \+ h1) -> 
        flats s1 \+ h1 = flats s2 \+ h2 ->
        graph_dec g s1 -> graph_dec g s2 ->
        s1 = s2.
Proof. by case: g=>g h V E [/= G1][->] C1 [_ /=][->]. Qed.

Lemma dec_inj g1 g2 s : 
        graph_dec g1 s -> graph_dec g2 s -> g1 = g2.
Proof. 
case:g1 g2=>g1 h1 [g2 h2][/= G1][-> C1][/= G2][E]_.
by case/scancelPt: E=>->->.
Qed.

Lemma dec_coh g s : graph_dec g s -> s \In Mod.coh W.
Proof. by case=>_[]. Qed.

Lemma dec_env g1 s1 : graph_dec g1 s1 -> env s1 = sig (Mod.coh W).
Proof. by case:g1=>g h[/=G][->]C; rewrite fsE. Qed.

Lemma dec_onto s2 : 
  s2 \In coh -> env s2 = sig (Mod.coh W) -> 
  exists g2, graph_dec g2 s2.
Proof.
move=>C E; case: (C)=>G[E']Vh Vso H.
exists (self s2, joint s2), G=>/=; split=>//.
suff G0 : other s2 = Unit; first by rewrite -G0. 
by rewrite -[s2]st_eta E /= getsoPt.
Qed.

End SpanDecorate.

Module Exports.
Notation graph_dec := graph_dec.

Section Exports.
Variable sp : nat.
Notation W := (SpanTree sp). Notation coh := (sp_coh sp).
Notation ptr_set := (ptrmap_pcm_Encoded [encoded_set of unit]).

Definition SpanDecorateMixin := 
  DecorateMixin (@dec_prec sp) (@dec_inj sp) (@dec_coh sp) (@dec_onto sp) (@dec_env sp).

Canonical SpanDecorate := 
  Eval hnf in Decorate (@graph_dec sp) SpanDecorateMixin.

End Exports.
End Exports.

End SpanDecorate.

Export SpanDecorate.Exports.

Section SpanHide.
Variables (pv sp : nat) (pf : sp != pv) (h : heap).
Notation W := (entangle (Priv pv) (SpanTree sp)).
Notation ptr_set := (ptrmap_pcm_Encoded [encoded_set of unit]).
Notation pv_getso s := (getso (pvg pv) s). Notation pv_self s := (self (pvg pv) s).
Notation pv_other s := (other (pvg pv) s).
Notation pv_coh := (priv_coh pv). Notation coh := (sp_coh sp).
Notation self s := (self (spg sp) s). Notation other s := (other (spg sp) s).
Notation joint s := (joint (spg sp) s). Notation subgraphT g1 g2 := (subgraph xpredT g1 g2).

Definition span_root_tp (x : ptr) :=
  {g1 : graph h}, STsep [entangle (Priv pv) empty_mod] 
    (fun s1 => (forall y, ~~mark g1 y) /\
               [/\ pv_self s1 = h, x \in dom h &
                   connected g1 x], 
   fun (_ : bool) s2 => exists g2 : graph (pv_self s2), 
     exists t,
     [/\ tree (edge g2) x t,
         forall x, (edgl g2 x \in [:: null; edgl g1 x]) &&
                   (edgr g2 x \in [:: null; edgr g1 x]) &
         dom t =i dom h]).
 
Program Definition span_root x : span_root_tp x := 
  Do (@priv_hide pv empty_mod (SpanTree sp) bool _ (SpanDecorate sp) (Unit, h) 
     (extend _ (span sp x))).
Next Obligation. exact: injRE (erefl _). Defined. 
Next Obligation.
apply: gh=>i1 g1 [Mk][Ep] D1 Cn1 C.
case: C Ep=>_ [i][_][->{i1}] C ->; rewrite {1}unitR.
case: C (C)=>_ [hL][hE] ->{i} /=; rewrite getE => C E; rewrite {hL}E in C *.
set st := sp ->> [Unit : ptr_set, h, Unit].
have Ct : st \In coh.
- rewrite -toPredE /= !getE; exists g1; split=>//.
  - by move: (cohH C); rewrite /validH priv_flatsE sp_flatsE; apply: validL.
  - by rewrite /validSO /st !fsE zipPt allvalidPt unitR valid_unit.
  by move=>z; rewrite unitR dom0 (negbTE (Mk z)).
rewrite -{1}[h]unitL {1}(_ : h = flats st); last by rewrite sp_flatsE.

(* Apply the hiding rule val_privhide'. *)

apply: val_privhide'; first by exists g1. 
- by rewrite unitR /st /valid !fsE /= !fvalidPtUn !fvalidPt !fdomPt pf.
rewrite joinC; apply: val_extend=>// {C}.
rewrite (_ : h = joint st) in g1 Mk D1 Cn1 *; last by rewrite /st getE.
apply: (gh_ex st); apply: (gh_ex g1); apply: val_do.
- by rewrite D1 orbT. 
rewrite getE=>b m [g2][S] /= X Cm j' Cmj' M i' x' hL' hE' [ns hm] E' d. 
case: d E'=>/= gn [->{x'}] Cn; rewrite sp_flatsE [_ \+ sp->>[_,_,_]]joinC=>E'.
move: (coh_prec (cohV Cmj') E' Cm Cn) => Em; rewrite Em in E' Cmj'.
case/(scancel (cohV Cmj')): E'=>_ V' E'.
rewrite getE; last by move: V'; rewrite E' /valid /= !fsE /= !fvalidPtUn. 
rewrite starC in Cmj'; move/(star_coh_prec Cmj'): Cn=>Cj'.
rewrite {j' V' Cmj'}E' in Cj' M.
have {Cj'} Cn': pv ->> [hL', tt, hE'] \In pv_coh.
- by split; [move/validHL: (cohH Cj') | exists hL', hE']. 
have Ck' : pv ->> [Unit, tt, hE] \In pv_coh.
- by case: (menvs_coh M); rewrite unitR /= starC0. 
case/(menvs_split (injLE _ (erefl _)) Ck' Cn'): M=>M _.  
rewrite (_ : hL' = Unit) ?unitR; last first.
- by move/menvs_loc: M; rewrite !fsE; move/(fcancelPt (C:=pcmCode)).
rewrite {m Cn' M}Em in g2 S X Cm; rewrite unitL; rewrite getE in X.
case: b X; last first.
- case=>N; rewrite {ns}N in g2 S Cm *.
  by rewrite (negbTE (dom_cond D1)) -sp_markE // !getE unitR dom0.
case: S=>D _ _ _ _ E [t][N]; rewrite {ns}N unitL in g2 D E Cm *.
rewrite !getE in g2 D E *; move: (@getPt (spg sp))=>/= L; rewrite {}L unitR. 
move=>_ T M F; exists g2, t; split=>// z; apply/idP/idP.
(* easy direction *)
- move=>H; rewrite D; apply: mark_dom. 
  rewrite (_ : hm = joint (sp ->> [t : ptr_set, hm, Unit])) in g2 E T M *.
  - by apply: markS=>//; rewrite getE.
  by rewrite getE.
(* there's a path from x to z because g1 is connected *)
case/Cn1=>p [P L]; apply: (max_path F P (proj1 T)).
by rewrite -L; apply: mem_last.
Qed.

End SpanHide.

